/* mm - Map mender. */
#include "../inc/ooGreedySub.inc"

double draftMinCoverage = 0.90;
double overlapMinCoverage = 0.995;

void usage()
/* Explain usage and exit. */
{
errAbort(
  "mm - Map mender\n"
  "usage:\n"
  "   mm testSet\n");
}

struct cloneEnd *newCloneEnd(struct oogClone *clone,  struct oogFrag *frag)
/* Allocate and initialize a clone end. */
{
struct cloneEnd *ce;
AllocVar(ce);
ce->clone = clone;
ce->frag = frag;
return ce;
}

void mmLoadClones(char *fileName, struct oogClone **retCloneList, struct hash **retCloneHash)
/* Load clones from file into list/hash. */
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *words[5];
struct oogClone *cloneList = NULL, *clone;
struct hash *cloneHash = newHash(11);
struct cloneEnd *ce;
int endCount;
static struct oogFrag dummyFrag;

while (lineFileRow(lf, words))
    {
    AllocVar(clone);
    slAddHead(&cloneList, clone);
    hashAddSaveName(cloneHash, words[0], clone, &clone->name);
    clone->phase = atoi(words[1]+1);
    clone->mapPos = lineFileNeedNum(lf, words, 2);
    endCount = lineFileNeedNum(lf, words, 3);
    clone->size = lineFileNeedNum(lf, words, 4);
    clone->sEnd = newCloneEnd(clone, (endCount > 0) ? &dummyFrag : NULL);
    clone->tEnd = newCloneEnd(clone, (endCount > 1) ? &dummyFrag : NULL);
    }
lineFileClose(&lf);
slReverse(&cloneList);
*retCloneList = cloneList;
*retCloneHash = cloneHash;
}

int ocpCmpScore(const void *va, const void *vb)
/* Compare to sort based on score. */
{
const struct overlappingClonePair *a = *((struct overlappingClonePair **)va);
const struct overlappingClonePair *b = *((struct overlappingClonePair **)vb);
return b->score - a->score;
}

void ocpDump(struct overlappingClonePair *ocp, FILE *f)
/* Print out ocp to file. */
{
fprintf(f, "%-12s %c %c     %-12s %c %c   %6d  %d %d\n",  
	ocp->a->name, ocp->aHitS, ocp->aHitT, 
	ocp->b->name, ocp->bHitS, ocp->bHitT, 
	ocp->overlap, ocp->externalPriority, ocp->score);
}

boolean ocpGreatEnds(char a, char b)
/* Returns TRUE if one of A is 'Y' and the other 'N' */
{
return (a == 'Y' && b == 'N') || (a == 'N' && b == 'Y');
}

boolean ocpHalfGood(char a, char b)
/* Returns TRUE if is of form ?N ?Y Y? N? */
{
return (a == '?' && (b == 'Y' || b == 'N')) || (b == '?' && (a == 'Y' || a == 'N'));
}

boolean ocpUnknownEnds(char a, char b)
/* Returns TRUE if is of form ?? */
{
return a == '?' && b == '?';
}

boolean ocpDoubleMiss(char a, char b)
/* Returns TRUE if is of form NN */
{
return a == 'N' && b == 'N';
}

boolean ocpDoubleHit(char a, char b)
/* RETURNS true if is of form YY */
{
return a == 'Y' && b == 'Y';
}

int ocpScore(struct overlappingClonePair *ocp)
/* Figure out score for ocp so that best looking overlaps have highest score. */
{
int score = 0;
int as = ocp->aHitS, at = ocp->aHitT, bs = ocp->bHitS, bt = ocp->bHitT;
struct oogClone *a = ocp->a, *b = ocp->b;
double size = max(a->size, b->size);
double dist;
struct oogClone *outerClone = NULL, *innerClone = NULL;

score += round(100000.0 * ocp->overlap / size);
score += ocp->externalPriority * 10000;
if (ocpDoubleMiss(as, at))
    {
    outerClone = a;
    innerClone = b;
    }
if (ocpDoubleMiss(bs, bt))
    {
    innerClone = b;
    outerClone = a;
    }
if (ocpGreatEnds(as, at) && ocpGreatEnds(bs, bt))
    score += 1000000;
else if (ocpGreatEnds(as, at) && ocpHalfGood(bs, bt))
    score += 500000;
else if (ocpGreatEnds(bs, bt) && ocpHalfGood(as, at))
    score += 500000;
else if (ocpHalfGood(as, at) && ocpHalfGood(bs, bt))
    score += 200000;
else if (ocpGreatEnds(as, at) && ocpUnknownEnds(bs, bt))
    score += 100000;
else if (ocpGreatEnds(bs, bt) && ocpUnknownEnds(as, at))
    score += 100000;
else if (ocpHalfGood(as, at) && ocpUnknownEnds(bs, bt))
    score += 50000;
else if (ocpHalfGood(bs, bt) && ocpUnknownEnds(as, at))
    score += 50000;
else if (outerClone != NULL)
    {
    int innerSize = innerClone->size;
    if (innerClone->phase != htgFinished)
        innerSize /= draftMinCoverage;
    if (ocp->overlap < innerSize)
        score -= 1000000;
    }
dist = a->mapPos - b->mapPos;
if (dist < 0) dist = -dist;
dist -= size;
if (dist > 600000)
    score -= 520000;
else if (dist > 300000)
    score -= (dist - 300000)/30;
return score;
}

void mmLoadOcp(char *fileName, struct hash *cloneHash,
	struct overlappingClonePair **retOcpList, struct hash **retOcpHash)
/* Load up overlaps from file into list/hash. */
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *words[8];
struct overlappingClonePair *ocpList = NULL, *ocp;
struct hash *ocpHash = newHash(12);

while (lineFileRow(lf, words))
    {
    AllocVar(ocp);
    slAddHead(&ocpList, ocp);
    hashAdd(ocpHash, ocpHashName(words[0], words[3]), ocp);
    ocp->a = hashMustFindVal(cloneHash, words[0]);
    ocp->b = hashMustFindVal(cloneHash, words[3]);
    ocp->overlap = lineFileNeedNum(lf, words, 6);
    ocp->aHitS = words[1][0];
    ocp->aHitT = words[2][0];
    ocp->bHitS = words[4][0];
    ocp->bHitT = words[5][0];
    ocp->externalPriority = lineFileNeedNum(lf, words, 7);
    ocp->score = ocpScore(ocp);
    }
lineFileClose(&lf);
slSort(&ocpList, ocpCmpScore);
*retOcpList = ocpList;
*retOcpHash = ocpHash;
}

void markBuriedClones(struct overlappingClonePair *ocpList)
/* Set buried field in bels. */
{
struct bargeEl *bel;
struct overlappingClonePair *ocp;
struct oogClone *clone;
int size;

for (ocp = ocpList; ocp != NULL; ocp = ocp->next)
    {
    if (ocp->a->size > ocp->b->size)
        {
	if (ocp->overlap >= overlapMinCoverage * ocp->b->size)
	    {
	    if (ocp->bHitS != 'N' && ocp->bHitT != 'N')
		{
		ocp->b->bel->enclosed = ocp->a->bel;
		fprintf(logFile, "%s covers %s\n", ocp->a->name, ocp->b->name);
		}
	    }
	}
    else
        {
	if (ocp->overlap >= overlapMinCoverage * ocp->a->size)
	    {
	    if (ocp->aHitS != 'N' && ocp->aHitT != 'N')
		{
		ocp->a->bel->enclosed = ocp->b->bel;
		fprintf(logFile, "%s covers %s\n", ocp->b->name, ocp->a->name);
		}
	    }
	}
    }
}

struct dlList *clonesToBels(struct oogClone *cloneList)
/* Create a bargeEl for each clone and put it on doubly-linked list. */
{
struct oogClone *clone;
struct bargeEl *bel;
struct dlNode *node;
struct dlList *list = newDlList();

for (clone = cloneList; clone != NULL; clone = clone->next)
    {
    AllocVar(bel);
    bel->clone = clone;
    clone->bel = bel;
    bel->dlNode = dlAddValTail(list, bel);
    }
return list;
}

boolean cloneHasKnownEnd(struct oogClone *clone)
/* Return TRUE if clone has a known end. */
{
return clone->sEnd->frag != NULL || clone->tEnd->frag != NULL;
}

struct oogClone *cloneFromBelNode(struct dlNode *node)
/* Return clone associated with bel associated with dlNode. */
{
struct bargeEl *bel = node->val;
return bel->clone;
}

int nextBargeId()
/* Return unique ID for a new barge. */
{
static int id = 100;
return id++;
}

struct barge *bargeOfOne(struct oogClone *clone, struct dlList *bargeList)
/* Construct a barge around a single clone. */
{
struct barge *barge;
struct cloneEnd *start, *end;
struct bargeEl *bel = clone->bel;

AllocVar(barge);
barge->id = nextBargeId();
barge->cloneEndList = newDlList();
barge->belList = newDlList();
dlAddValTail(barge->belList, bel);
barge->dlNode = dlAddValTail(bargeList, barge);
start = CloneVar(clone->sEnd);
start->isStart = TRUE;
dlAddValTail(barge->cloneEndList, start);
end = CloneVar(clone->tEnd);
end->isStart = FALSE;
if (clone->sEnd->frag || clone->tEnd->frag)
    start->orientation = end->orientation = 1;
dlAddValTail(barge->cloneEndList, end);
clone->barge = barge;
return barge;
}

void findCloneEndsOnList(struct dlList *list, struct oogClone *clone, 
	struct dlNode **retStart, struct dlNode **retEnd)
/* Find start and end nodes associated with clone on list. */
{
struct dlNode *node;
struct cloneEnd *ce;
boolean gotStart = FALSE, gotEnd = FALSE;
for (node = list->head; !dlEnd(node); node = node->next)
     {
     ce = node->val;
     if (ce->clone == clone)
         {
	 if (!gotStart)
	     {
	     assert(ce->isStart);
	     gotStart = TRUE;
	     *retStart = node;
	     }
	 else
	     {
	     assert(!ce->isStart);
	     gotEnd = TRUE;
	     *retEnd = node;
	     break;
	     }
         }
     }
if (!gotEnd)
    errAbort("Couldn't find ends of %s in list", clone->name);
}

char ocpHitsHow(struct overlappingClonePair *ocp, struct oogClone *clone, 
	int relPos)
/* Return how other clone in pair hits clone assuming other end is oriented. 
 * RelPos is whether clone is to left (+1) or right (-1) of other. */
{
char sHit, tHit;
if (clone == ocp->a)
    {
    sHit = ocp->aHitS;
    tHit = ocp->aHitT;
    }
else
    {
    assert(clone == ocp->b);
    sHit = ocp->bHitS;
    tHit = ocp->bHitT;
    }
if (relPos > 0)
    return tHit;
else if (relPos < 0)
    return sHit;
else
    return '?';
}

struct fitAt
/* A possible position clone fits. */
    {
    struct fitAt *next;
    struct dlNode *start, *end;		/* cloneEnd list nodes where clone goes. */
    int orientation;			/* +1, -1 or 0. */
    boolean addAfter;			/* True if add clone after start/end. */
    int score;				/* Some sort of score. */
    };

void dumpFit(struct fitAt *fit, struct oogClone *newbie)
/* Print out info on fit. */
{
struct cloneEnd *se = fit->start->val, *ee = fit->end->val;
uglyf(" %s fits %s %s (%s) %s (%s), score %d\n",
    newbie->name, 
    fit->addAfter ? "after" : "before",
    se->clone->name,
    se->isStart ? "start" : "end",
    ee->clone->name,
    ee->isStart ? "start" : "end",
    fit->score);
}

int fitAtCmpScore(const void *va, const void *vb)
/* Compare to sort based on score. */
{
const struct fitAt *a = *((struct fitAt **)va);
const struct fitAt *b = *((struct fitAt **)vb);
return b->score - a->score;
}

int overlapSize(struct oogClone *a, struct oogClone *b, struct hash *ocpHash)
/* Return size of overlap between a and b. */
{
struct overlappingClonePair *ocp;

if (a == b)
    return a->size;
ocp = findHashedOcp(a->name, b->name, ocpHash);
if (ocp == NULL)
    return 0;
return ocp->overlap;
}

int addMissingOverlap(struct dlNode *startN, struct dlNode *endN, 
	struct oogClone *newbie, struct hash *ocpHash)
/* Add together all overlaps that should be present between
 * newbie and clones, assuming that newbie covers all of the
 * ends in between startN and endN (including startN and endN) 
 * The result will be 0 if all overlaps are present, or a negative
 * number if some are missed. 
 *    This routine assumes that there are no clones fully enclosed
 * between startN and endN. */
{
struct cloneEnd *startE = startN->val, *endE = endN->val, *ce;
struct oogClone *startC = startE->clone, *endC = endE->clone, *clone;
int missedOverlap = 0, missedOne, minOverlap, overlap;
int presentOverlap = 0;
struct dlNode *node;

for (node = startN; node != endN->next; node = node->next)
    {
    ce = node->val;
    clone = ce->clone;
    if (ce->isStart)
        {
	if (endE->isStart)
	    minOverlap = clone->size - overlapSize(clone, endC, ocpHash);
	else
	    minOverlap = overlapSize(clone, endC, ocpHash);
	}
    else
        {
	if (startE->isStart)
	    minOverlap =  overlapSize(clone, startC, ocpHash);
	else
	    minOverlap = clone->size - overlapSize(clone, startC, ocpHash);
	}
    overlap = overlapSize(clone, newbie, ocpHash);
    presentOverlap += overlap;
    missedOne = overlap - minOverlap;
    if (missedOne < 0)
        missedOverlap += missedOne;
    }
if (missedOverlap < 0)
    return missedOverlap;
else
    return presentOverlap;
}

boolean insideEndsFit(struct oogClone *newbie, 
	struct dlNode *innerStart, struct dlNode *outerEnd, 
	struct hash *ocpHash)
/* Return TRUE if all known ends from innerStart to outerEnd (half open interval)
 * hit newbie, and that newbie overlaps with all clones in interval. */
{
struct dlNode *n1;
struct oogClone *clone;
struct cloneEnd *ce;
struct overlappingClonePair *ocp;
int lorient;
char how;

uglyf("      ---insideEndsFit: newbie %s\n", newbie->name);

/* See if this implies overlaps that don't exist. */
for (n1 = innerStart; n1 != outerEnd; n1 = n1->next)
    {
    ce = n1->val;
    clone = ce->clone;
    uglyf("       --- %s (%s): ", clone->name, (ce->isStart ? "start" : "end"));
    ocp = findHashedOcp(newbie->name, clone->name, ocpHash);
    if (ocp == NULL)
	{
	uglyf(" no overlap\n");
	return FALSE;
	}

	{
	ocpDump(ocp, logFile);
	ocpDump(ocp, uglyOut);
	}
    /* Check end specific matches. */
    lorient = ce->orientation;
    if (ce->isStart)
	lorient = -lorient;
    how = ocpHitsHow(ocp, clone, lorient);
    if (how == 'N')
	return FALSE;
    }
return TRUE;
}

struct fitAt *cloneFitsAt(struct oogClone *newbie, struct dlNode *startN, struct dlNode *endN,
	boolean addAfter, int newbieOrientation, 
	struct oogClone *member, struct dlNode *memStart, 
	struct dlNode *memEnd, struct hash *ocpHash)
/* Return a fitAt structure if clone can fit at position . */
{
struct dlNode *n1, *n2;
struct dlNode *innerStart, *innerEnd;
struct cloneEnd *ce;
struct oogClone *clone;
struct overlappingClonePair *ocp;
char how;
struct fitAt *fit;
int score, minScore;


/* Figure out adjStart/End which define the first end
 * inside of newbie, and the first end after newbie. */
if (addAfter)
    {
    innerStart = startN->next;
    innerEnd = endN->next;
    }
else
    {
    innerStart = startN;
    innerEnd = endN;
    }

/* See if this would lead to new clone enclosing another. */
for (n1 = innerStart; n1 != innerEnd; n1 = n1->next)
    {
    ce = n1->val;
    clone = ce->clone;
    if (ce->isStart)
        {
	for (n2 = n1->next; n2 != innerEnd; n2 = n2->next)
	    {
	    ce = n2->val;
	    if (ce->clone == clone)
	        {
		return NULL;
		}
	    }
	}
    }


/* See if this would lead to new clone being enclosed. */
if (addAfter)
    {
    for (n1 = memStart; n1 != memEnd; n1 = n1->next)
	{
	boolean newbieStartsInside = FALSE, newbieEndsInside = FALSE;
	ce = n1->val;
	clone = ce->clone;
	if (ce->isStart)
	    {
	    for (n2 = n1; ;n2 = n2->next)
		{
		ce = n2->val;
		if (n2 == startN)
		    newbieStartsInside = TRUE;
		if (ce->clone == clone && !ce->isStart)
		    break;
		if (n2 == endN)
		    newbieEndsInside = TRUE;
		}
	    }
	else
	    {
	    for (n2 = n1->prev; ;n2 = n2->prev)
		{
		if (n2 == endN)
		    newbieEndsInside = TRUE;
		if (n2 == startN)
		    newbieStartsInside = TRUE;
		ce = n2->val;
		if (ce->clone == clone && ce->isStart)
		    break;
		}
	    }
	if (newbieStartsInside && newbieEndsInside)
	    return NULL;
	}
    }
else
    {
    for (n1 = memEnd; n1 != memStart; n1 = n1->prev)
	{
	boolean newbieStartsInside = FALSE, newbieEndsInside = FALSE;
	ce = n1->val;
	clone = ce->clone;
	if (!ce->isStart)
	    {
	    for (n2 = n1; ;n2 = n2->prev)
		{
		ce = n2->val;
		if (n2 == endN)
		    newbieEndsInside = TRUE;
		if (ce->clone == clone && ce->isStart)
		    break;
		if (n2 == startN)
		    newbieStartsInside = TRUE;
		}
	    }
	else
	    {
	    for (n2 = n1->next; ;n2 = n2->next)
		{
		if (n2 == startN)
		    newbieStartsInside = TRUE;
		if (n2 == endN)
		    newbieEndsInside = TRUE;
		ce = n2->val;
		if (ce->clone == clone && !ce->isStart)
		    break;
		}
	    }
	if (newbieStartsInside && newbieEndsInside)
	    return NULL;
	}
    }


if (!insideEndsFit(newbie, innerStart, innerEnd, ocpHash))
    return NULL;

/* Score fit according to how well overlap distances add up. */
if (addAfter)
    score = addMissingOverlap(startN->next, endN, newbie, ocpHash);
else
    score = addMissingOverlap(startN, endN->prev, newbie, ocpHash);

minScore = newbie->size/8;
if (newbie->phase == htgFinished)
    minScore /= 2;
if (score < -minScore)
    return NULL;

/* Allocate and fill in fit structure. */
AllocVar(fit);
fit->start = startN;
fit->end = endN;
fit->orientation = newbieOrientation;
fit->addAfter = addAfter;
fit->score = score;

dumpFit(fit, newbie);
return fit;
}

void addFittedClone(struct barge *barge, struct oogClone *clone, struct fitAt *fit)
/* Add clone at specific position in barge. */
{
struct dlNode *(*listAdder)(struct dlNode *anchor, void *val);
struct dlNode *firstNode;
struct cloneEnd *start,*end;

listAdder = (fit->addAfter ? dlAddValAfter : dlAddValBefore);
dlAddValTail(barge->belList, clone->bel);
clone->barge = barge;
if (fit->orientation < 0)
    {
    start = clone->tEnd;
    end = clone->sEnd;
    }
else
    {
    start = clone->sEnd;
    end = clone->tEnd;
    }
start = CloneVar(start);
end = CloneVar(end);
start->isStart = TRUE;
end->isStart = FALSE;
start->orientation = end->orientation = fit->orientation;
firstNode = (*listAdder)(fit->start, start);
if (fit->end != fit->start)
    (*listAdder)(fit->end, end);
else
    dlAddValAfter(firstNode, end);
}

enum orientTypes 
/* Possible clone orientations. */
  {
  orientImpossible = -2,
  orientForward = 1,
  orientReverse = -1,
  orientUnknown = 0,
  };

enum orientTypes ocpCloneOrientation(struct overlappingClonePair *ocp, struct oogClone *clone)
/* Return orientation of clone relative to other clone in pair based on end matches. 
 * Returns 1 (orientForward) if it appears clone comes after other clone. */
{
int orientation = orientUnknown;
char how, howFlip;

how = ocpHitsHow(ocp, clone, -1);
howFlip = ocpHitsHow(ocp, clone, 1);
if (how == 'N' && howFlip == 'N')
    return orientImpossible;
if (how == 'Y')
    return orientForward;
else if (howFlip == 'Y')
    return orientReverse;
else if (how == 'N')
    return orientReverse;
else if (howFlip == 'N')
    return orientForward;
else if (howFlip == '?' && how == '?')
    return orientUnknown;
else
    {
    warn("Case that should never happen in ocpCloneOrientation.");
    assert(FALSE);
    return orientImpossible;
    }
}

boolean bargeAddLeft(struct barge *barge, 
	struct oogClone *newbie,  struct oogClone *member,
	struct dlNode *memStart, struct dlNode *memEnd, 
	struct overlappingClonePair *joinOcp,
	struct hash *ocpHash, boolean doAdd,
	int *retOrientation)
/* See if can add barge to left of member (which is defined by it's ends on
 * barge->cloneEndList) */
{
char how, howFlip, newHow;
int newbieOrientation = 0, memberOrientation;
struct dlNode *node, *firstStart, *n1, *n2;
struct cloneEnd *ce;
struct oogClone *clone;
struct overlappingClonePair *ocp;
struct fitAt *fitList = NULL, *fit;
int score;

ce = memStart->val;
memberOrientation = ce->orientation;

how = ocpHitsHow(joinOcp, member, -memberOrientation);
if (how == 'N')
    return FALSE;
howFlip = ocpHitsHow(joinOcp, member, memberOrientation);
if (howFlip == 'Y')
    return FALSE;	/* Might eventually want to just decrease score greatly. */
if (cloneHasKnownEnd(newbie))
    {
    newbieOrientation = ocpCloneOrientation(joinOcp, newbie);
    if (newbieOrientation == orientImpossible)
        return FALSE;
    newbieOrientation = -newbieOrientation;
    }


/* Figure out where can stop scanning for newbie start sites
 * because have a clone ending that newbie doesn't overlap. */
for (node = memStart->prev; !dlStart(node); node = node->prev)
    {
    ce = node->val;
    if (!ce->isStart)
        {
	clone = ce->clone;
	ocp = findHashedOcp(clone->name, newbie->name, ocpHash);
	if (ocp == NULL)
	    break;
	}
    }
firstStart = node;

for (n1 = memEnd; n1 != memStart; n1 = n1->prev)
    {
    for (n2 = memStart; n2 != firstStart; n2 = n2->prev)
        {
	if ((fit = cloneFitsAt(newbie, n2, n1, FALSE, 
		newbieOrientation, member, memStart, memEnd, ocpHash)) != NULL)
	    {
	    slAddHead(&fitList, fit);
	    }
	}
    }
slSort(&fitList, fitAtCmpScore);
if ((fit = fitList) != NULL)
    {
    if (doAdd)
        addFittedClone(barge, newbie, fit);
    *retOrientation = fit->orientation;
    }
slFreeList(&fitList);
return fit != NULL;
}

boolean bargeAddRight(struct barge *barge, 
	struct oogClone *newbie,  struct oogClone *member,
	struct dlNode *memStart, struct dlNode *memEnd, 
	struct overlappingClonePair *joinOcp,
	struct hash *ocpHash, boolean doAdd,
	int *retOrientation)
/* See if can add barge to right of member (which is defined by it's ends on
 * barge->cloneEndList) */
{
char how, howFlip, newHow;
int newbieOrientation = 0, memberOrientation ;
struct dlNode *node, *lastEnd, *n1, *n2;
struct cloneEnd *ce;
struct oogClone *clone;
struct overlappingClonePair *ocp;
struct fitAt *fitList = NULL, *fit;
int score;

ce = memStart->val;
memberOrientation = ce->orientation;
how = ocpHitsHow(joinOcp, member, memberOrientation);
if (how == 'N')
    return FALSE;
howFlip = ocpHitsHow(joinOcp, member, -memberOrientation);
if (howFlip == 'Y')
    return FALSE;	/* Might eventually want to just decrease score greatly. */
if (cloneHasKnownEnd(newbie))
    {
    newbieOrientation = ocpCloneOrientation(joinOcp, newbie);
    if (newbieOrientation == orientImpossible)
        return FALSE;
    }


/* Figure out where can stop scanning for newbie end sites
 * because have a clone starting that newbie doesn't overlap. */
for (node = memEnd->next; !dlEnd(node); node = node->next)
    {
    ce = node->val;
    if (ce->isStart)
        {
	clone = ce->clone;
	ocp = findHashedOcp(clone->name, newbie->name, ocpHash);
	if (ocp == NULL)
	    break;
	}
    }
lastEnd = node;

for (n1 = memStart; n1 != memEnd; n1 = n1->next)
    {
    for (n2 = memEnd; n2 != lastEnd; n2 = n2->next)
        {
	if ((fit = cloneFitsAt(newbie, n1, n2, TRUE, 
		newbieOrientation, member, memStart, memEnd, ocpHash)) != NULL)
	    {
	    slAddHead(&fitList, fit);
	    }
	}
    }
slSort(&fitList, fitAtCmpScore);
if ((fit = fitList) != NULL)
    {
    if (doAdd)
        addFittedClone(barge, newbie, fit);
    *retOrientation = fit->orientation;
    }
slFreeList(&fitList);
return fit != NULL;
}

void dumpCloneEnds(struct dlList *list, FILE *f)
/* Dump out info on clone ends in list to f. */
{
struct dlNode *node;
int orientation;

for (node = list->head; !dlEnd(node); node = node->next)
    {
    struct cloneEnd *ce = node->val;
    orientation = ce->orientation;
    fprintf(f, " ");
    if (ce->isStart)
	{
	if (orientation > 0)
	    fprintf(f, "+");
	else if (orientation < 0)
	    fprintf(f, "-");
	else
	    fprintf(f, "?");
        fprintf(f, "(");
	}
    fprintf(f, "%s", ce->clone->name);
    if (!ce->isStart)
        fprintf(f, ")");
    }
fprintf(f, "\n");
}

boolean bargeAdd(struct barge *barge, struct oogClone *newbie, struct oogClone *member,
	struct hash *ocpHash, boolean doAdd)
/* Attempt to add a clone 'newbie' to barge via connection with 'member'. */
{
struct dlNode *memStart, *memEnd;	/* Start/end nodes of member. */
struct overlappingClonePair *ocp;
int ori;

uglyf("bargeAdd: newbie %s, member %s\n", newbie->name, member->name);
dumpCloneEnds(barge->cloneEndList, uglyOut);
dumpCloneEnds(barge->cloneEndList, logFile);
ocp = findHashedOcp(newbie->name, member->name, ocpHash);
findCloneEndsOnList(barge->cloneEndList, member, &memStart, &memEnd);
if (bargeAddRight(barge, newbie, member, memStart, memEnd, ocp, ocpHash, doAdd, &ori))
    {
    dumpCloneEnds(barge->cloneEndList, uglyOut);
    dumpCloneEnds(barge->cloneEndList, logFile);
    return TRUE;
    }
else
    {
    uglyf(" not on right\n");
    }
if (bargeAddLeft(barge, newbie, member, memStart, memEnd, ocp, ocpHash, doAdd, &ori))
    {
    dumpCloneEnds(barge->cloneEndList, uglyOut);
    dumpCloneEnds(barge->cloneEndList, logFile);
    return TRUE;
    }
else
    {
    uglyf(" not on left\n");
    return FALSE;
    }
}

struct barge *bargeOfTwo(struct overlappingClonePair *ocp, struct dlList *bargeList,
	struct hash *ocpHash, struct overlappingClonePair *ocpList)
/* Make up a barge with just two elements. */
{
struct oogClone *cloneA = ocp->a, *cloneB = ocp->b;
struct bargeEl *belA = cloneA->bel, *belB = cloneB->bel;
struct barge *barge;

uglyf("Barge of two: %s %s %d %d\n", cloneA->name, cloneB->name, ocp->overlap, ocp->score);

barge = bargeOfOne(cloneA, bargeList);
if (!bargeAdd(barge, cloneB, cloneA, ocpHash, TRUE))
    {
    //errAbort("Can't add second barge in bargeOfTwo(%s %s)", cloneA->name, cloneB->name);
    warn("Can't add second barge in bargeOfTwo(%s %s)", cloneA->name, cloneB->name);
    cloneA->barge = NULL;	/* uglyf */
    }
return barge;
}

void addToBarge(struct barge *barge, struct overlappingClonePair *ocp,
	struct oogClone *member, struct oogClone *newbie, 
	struct hash *ocpHash, struct overlappingClonePair *ocpList)
/* Add clone to existing barge. */
{
uglyf("Adding clone %s to barge with %s\n", newbie->name, member->name);
bargeAdd(barge, newbie, member, ocpHash, TRUE);
}

struct dlList *dlClone(struct dlList *orig)
/* Make a copy of orig list and return it. */
{
struct dlList *cc = newDlList();
struct dlNode *node;
for (node = orig->head; !dlEnd(node); node = node->next)
    dlAddValTail(cc, node->val);
return cc;
}

/* When merging barges we do it tentatively, first saving
 * enough state to restore data structures if merger doesn't
 * work out, and then merging.  */

boolean tentativelyMergeBarges(struct barge *big, struct barge *little, 
    struct overlappingClonePair *joiningOcp, struct hash *ocpHash)
/* Merge little barge into big barge.  If there's a problem
 * return false.  The big barge may have the little barge
 * partly merged in at this stage.  The function "mergeBarge"
 * below will clean things up if this happens. */
{
struct oogClone *member, *newbie, *initialNewbie;
int littleOrientation, bigOrientation;
int relBargeOrientation = 0;
struct dlNode *iNewStart, *iNewEnd;
struct dlNode *memStart, *memEnd;
struct dlNode *n1;
struct cloneEnd *ce;
boolean leftOk, rightOk;
struct overlappingClonePair *ocp;
char *newbieName;

/* Figure out which clone is already part of big barge. */
if (joiningOcp->a->barge == big)
    {
    member = joiningOcp->a;
    initialNewbie = joiningOcp->b;
    }
else
    {
    assert(joiningOcp->b->barge == big);
    member = joiningOcp->b;
    initialNewbie = joiningOcp->a;
    }

uglyf("Tentatively merging barges\n");
uglyf("big: ");
dumpCloneEnds(big->cloneEndList, uglyOut);
dumpCloneEnds(big->cloneEndList, logFile);
uglyf("little: ");
dumpCloneEnds(little->cloneEndList, uglyOut);
dumpCloneEnds(little->cloneEndList, logFile);

findCloneEndsOnList(little->cloneEndList, initialNewbie, &iNewStart, &iNewEnd);
for (n1 = iNewStart; !dlEnd(n1); n1 = n1->next)
    {
    ce = n1->val;
    if (ce->isStart)
        {
	newbie = ce->clone;
	leftOk = rightOk = FALSE;
	littleOrientation = ce->orientation;
	findCloneEndsOnList(big->cloneEndList, member, &memStart, &memEnd);
	ocp = findHashedOcp(member->name, newbie->name, ocpHash);
	if (ocp == NULL)
	    errAbort("Buh, neighbors don't overlap in merge barge1 ?");
	if (relBargeOrientation >= 0)
	    {
	    rightOk = bargeAddRight(big, newbie, member, memStart, memEnd, ocp, 
	    	ocpHash, TRUE, &bigOrientation);
	    if (rightOk)
		{
		if (newbie != initialNewbie)
		    relBargeOrientation = 1;
		}
	    else 
	        {
		if (relBargeOrientation != 0)
		    {
		    uglyf("Couldn't tentatively merge right\n");
		    return FALSE;
		    }
		}
	    }
	if (relBargeOrientation <= 0 && !rightOk)
	    {
	    leftOk = bargeAddLeft(big, newbie, member, memStart, 
	    	memEnd, ocp, ocpHash, TRUE, &bigOrientation);
	    if (leftOk)
		{
		if (newbie != initialNewbie)
		    relBargeOrientation = -1;
		}
	    else
	        {
		if (relBargeOrientation == 0)
		    uglyf("Couldn't tentatively merge initial\n");
		else
		    uglyf("Couldn't tentatively merge left\n");
		return FALSE;
		}
	    }
	if (bigOrientation != 0 && littleOrientation != 0 && relBargeOrientation != 0)
	    {
	    if (relBargeOrientation != bigOrientation*littleOrientation)
	        errAbort("Inconsistent orientations in tentatively merge");
	    }
	dumpCloneEnds(big->cloneEndList, uglyOut);
	dumpCloneEnds(big->cloneEndList, logFile);
	}
    member = newbie;
    }
member = initialNewbie;
for (n1 = iNewStart->prev; !dlStart(n1); n1 = n1->prev)
    {
    ce = n1->val;
    if (ce->isStart)
        {
	newbie = ce->clone;
	littleOrientation = ce->orientation;
	findCloneEndsOnList(big->cloneEndList, member, &memStart, &memEnd);
	ocp = findHashedOcp(member->name, newbie->name, ocpHash);
	if (ocp == NULL)
	    errAbort("Buh, neighbors don't overlap in merge barge2 ?");
	if (relBargeOrientation <= 0)
	    {
	    rightOk = bargeAddRight(big, newbie, member, memStart, memEnd, ocp, 
	    	ocpHash, TRUE, &bigOrientation);
	    if (rightOk)
		{
		relBargeOrientation = -1;
		}
	    else 
	        {
		if (relBargeOrientation != 0)
		    {
		    uglyf("Couldn't tentatively merge right 2\n");
		    return FALSE;
		    }
		}
	    }
	if (relBargeOrientation >= 0)
	    {
	    leftOk = bargeAddLeft(big, newbie, member, memStart, memEnd, ocp, 
	    	ocpHash, TRUE, &bigOrientation);
	    if (leftOk)
		{
		relBargeOrientation = 1;
		}
	    else
	        {
		uglyf("Couldn't tentatively merge left 2\n");
		return FALSE;
		}
	    }
	if (bigOrientation != 0 && littleOrientation != 0 && relBargeOrientation != 0)
	    {
	    if (relBargeOrientation != bigOrientation*littleOrientation)
	        errAbort("Inconsistent orientations in tentatively merge 2");
	    }
	dumpCloneEnds(big->cloneEndList, uglyOut);
	dumpCloneEnds(big->cloneEndList, logFile);
	}
    member = newbie;
    }
return TRUE;
}

struct barge *backupBarge(struct barge *orig)
/* Make a lightweight copy of barge - just enough to restore
 * barge if need be. */
{
struct barge *cc;
AllocVar(cc);
cc->cloneEndList = dlClone(orig->cloneEndList);
cc->belList = dlClone(orig->belList);
return cc;
}

void reassertBargeCloneOwnership(struct barge *barge)
/* Make sure all clones part of barge point to barge. */
{
struct dlNode *node;
struct bargeEl *bel;
for (node = barge->belList->head; !dlEnd(node); node = node->next)
    {
    bel = node->val;
    bel->clone->barge = barge;
    }
}

void fixIsStarts(struct dlList *cloneEndList)
/* Fix up isStart fields on clone ends in list. */
{
struct hash *hash = newHash(8);
struct dlNode *node;
struct cloneEnd *ce;
char *cloneName;

uglyf("Fixing IsStarts\n");
uglyf("before: ");
dumpCloneEnds(cloneEndList, uglyOut);
dumpCloneEnds(cloneEndList, logFile);
for (node = cloneEndList->head; !dlEnd(node); node = node->next)
    {
    ce = node->val;
    cloneName = ce->clone->name;
    if (hashLookup(hash, cloneName))
        ce->isStart = FALSE;
    else
        {
	ce->isStart = TRUE;
	hashAdd(hash, cloneName, NULL);
	}
    }
freeHash(&hash);
uglyf("after: ");
dumpCloneEnds(cloneEndList, uglyOut);
dumpCloneEnds(cloneEndList, logFile);
}

void restoreBarge(struct barge *barge, struct barge *backup)
/* Restore barge from backup. */
{
freeDlListAndVals(&barge->cloneEndList);
freeDlList(&barge->belList);
barge->cloneEndList = backup->cloneEndList;
backup->cloneEndList = NULL;
barge->belList = backup->belList;
backup->belList = NULL;
}


boolean mergeBarges(struct barge *bargeA, struct barge *bargeB, 
	struct overlappingClonePair *ocp, struct hash *ocpHash,
	struct dlList *bargeList)
/* Attempt to merge together two barges.  If can't restore things
 * so that original barges unchanged. */
{
struct barge *big, *little, *backup;
int aCount = dlCount(bargeA->belList), bCount = dlCount(bargeB->belList);
boolean ok;

uglyf("Merging barges based on overlap between %s and %s\n", ocp->a->name, ocp->b->name);
uglyf("BargeA has %d clones, bargeB has %d clones\n", aCount, bCount);
if (aCount >= bCount)
    {
    big = bargeA;
    little = bargeB;
    }
else
    {
    big = bargeB;
    little = bargeA;
    }
backup = backupBarge(big);
ok = tentativelyMergeBarges(big, little, ocp, ocpHash);
if (!ok)
    {
    restoreBarge(big, backup);
    reassertBargeCloneOwnership(little);
    fixIsStarts(little->cloneEndList);
    }
else
    {
    dlRemove(little->dlNode);
    freez(&little->dlNode);
    freeDlList(&little->cloneEndList);
    freeDlList(&little->belList);
    freez(&little);
    }
freeDlList(&backup->cloneEndList);
freeDlList(&backup->belList);
freez(&backup);
return ok;
}

int scoreEnclosingOverlap(struct overlappingClonePair *ocp, struct oogClone *inner)
/* Score whether it looks like outer clone encloses inner clone.
 * Return number between -2 (no way!) to 1 (yep) */
{
int innerSize = inner->size;
int minEncloseSize = inner->size  * overlapMinCoverage;

if (ocp == NULL)
    return -2;
uglyf("        scoreEnclosingOverlap %s&%s overlap %d, inner %s, innerSize %d, minEncloseSize %d\n", ocp->a->name, ocp->b->name, ocp->overlap,  inner->name, inner->size, minEncloseSize);
if (ocp->overlap <= innerSize * 0.75)
    return -1;
if (ocp->overlap <= minEncloseSize)
    return 0;
return 1;
}

int scoreNonenclosingOverlap(struct overlappingClonePair *ocp, struct oogClone *a, struct oogClone *b)
/* Score whether it looks like two clones overlap, but neither is enclosed. 
 * Returns 1 (yes), -1 (no) or 0 (maybe). */
{
if (ocp == NULL)
    return -1;
if (ocpGreatEnds(ocp->aHitS, ocp->aHitT))
    return 1;
if (ocpGreatEnds(ocp->bHitS, ocp->bHitT))
    return 1;
if (ocp->overlap >= a->size * overlapMinCoverage)
    return 0;
if (ocp->overlap >= b->size * overlapMinCoverage)
    return 0;
return 1;
}

boolean orientPair(struct overlappingClonePair *ocp, struct oogClone *clone,
	int relOrientation, int *pCloneOrientation)
/* See if orientation of clone either is, or if unknown can be made
 * consistent with position implied by one clone coming after the other 
 * and ocp. Just checks clone ends, not other clone ends
 * if any in ocp. */
{
int oldOrientation = *pCloneOrientation;
int ocpOrientation = ocpCloneOrientation(ocp, clone);

if (ocpOrientation == orientImpossible)
    return FALSE;
if (ocpDoubleMiss(ocp->aHitS, ocp->aHitT))
    return FALSE;
if (ocpDoubleMiss(ocp->bHitS, ocp->bHitT))
    return FALSE;
uglyf("        orientPair: ");
ocpDump(ocp, logFile);
ocpDump(ocp, uglyOut);
uglyf("        %s of %s-%s: oldOrientation %d, ocpOrientation %d,  relOrientation %d\n",
    clone->name, ocp->a->name, ocp->b->name, oldOrientation, ocpOrientation, relOrientation);
if (ocpOrientation != orientUnknown)
    {
    ocpOrientation *= relOrientation;
    if (oldOrientation != orientUnknown && oldOrientation != ocpOrientation)
        return FALSE;   /* Contradicts earlier orientation. */
    *pCloneOrientation = ocpOrientation;
    }
return TRUE;
}

struct fitAt *fitEnclosed(struct dlNode *listStart, struct dlNode *listEnd, 
	struct oogClone *clone, struct dlNode *outerStart, struct dlNode *outerEnd,
	struct hash *ocpHash)
/* Figure out if can place clone between outerStart and outerEnd. */
{
struct dlNode *innerStart = outerStart->next, *innerEnd = outerEnd->prev;
struct dlNode *testStart, *testEnd;
struct cloneEnd *ce;
struct oogClone *testClone;
boolean pastCloneStart = FALSE;
int fitScore = 0, oneScore;
struct overlappingClonePair *ocp;
struct fitAt *fit;
int newbieOrientation = 0;

    {
    struct cloneEnd *ce1 = outerStart->val, *ce2 = outerEnd->val;
    char name1[64], name2[64];
    if (ce1 == NULL)
          sprintf(name1, "BEGIN");
    else
          sprintf(name1, "%s(%s)", ce1->clone->name, 
	      (ce1->isStart ? "start" : "end"));
    if (ce2 == NULL)
          sprintf(name2, "END");
    else
          sprintf(name2, "%s(%s)", ce2->clone->name, 
	      (ce2->isStart ? "start" : "end"));
    uglyf("   fitEnclosed( %s between %s and %s)\n", clone->name, name1, name2);
    }

if (!insideEndsFit(clone, innerStart, outerEnd, ocpHash))
    {
    uglyf("      insideEnds don't fit\n");
    return NULL;
    }

for (testStart = listStart; testStart != listEnd; testStart = testStart->next)
    {
    if (testStart == innerStart) pastCloneStart = TRUE;
    ce = testStart->val;
    if (ce->isStart)
        {
	struct dlNode *gotInStart, *gotInEnd, *gotOutStart,*gotOutEnd;
	gotInStart = gotInEnd = gotOutStart = gotOutEnd = NULL;
	testClone = ce->clone;
	for (testEnd = testStart;  ;testEnd = testEnd->next)
	    {
	    if (testEnd == innerStart) gotInStart = testEnd;
	    if (testEnd == outerStart) gotOutStart = testEnd;
	    if (testEnd == innerEnd) gotInEnd = testEnd;
	    if (testEnd == outerEnd) gotOutEnd = testEnd;
	    ce = testEnd->val;
	    if (ce->clone == testClone && !ce->isStart)
		break;
	    }
	ocp = findHashedOcp(testClone->name, clone->name, ocpHash);
	if (gotOutStart != NULL && gotOutEnd!= NULL )
	     {
	     /* Case where test encloses clone. */
	     oneScore = scoreEnclosingOverlap(ocp, clone);
	     fitScore += oneScore;
	     uglyf("       %s encloses %s score %d\n", testClone->name, clone->name, oneScore);
	     }
	else if (gotOutStart == NULL && gotOutEnd == NULL)
	     {
	     if (pastCloneStart && innerStart != outerEnd)
		 {
		 /* Case where clone encloses test. */
		 oneScore = scoreEnclosingOverlap(ocp, testClone);
		 fitScore += oneScore;
		 uglyf("       %s enclosed by %s score %d\n", testClone->name, clone->name, oneScore);
		 }
	     else
	         {
		 /* Case where test is before clone, no overlap. */
		 }
	     }
	else if (gotOutStart != NULL && gotOutStart != testEnd)
	     {
	     /* Case where order is testClone, clone and the two overlap. */
	     oneScore = scoreNonenclosingOverlap(ocp, clone, testClone);
	     if (!orientPair(ocp, clone, 1, &newbieOrientation))
	         return NULL;
	     fitScore += oneScore;
	     uglyf("       %s before %s score %d\n", testClone->name, clone->name, oneScore);
	     }
	else if (gotOutEnd != NULL && gotOutEnd != testStart)
	     {
	     /* Case where order is clone, testClone and the two overlap. */
	     oneScore = scoreNonenclosingOverlap(ocp, testClone, clone);
	     fitScore += oneScore;
	     if (!orientPair(ocp, clone, -1, &newbieOrientation))
	         return NULL;
	     uglyf("       %s after %s score %d\n", testClone->name, clone->name, oneScore);
	     }
	}
    }
AllocVar(fit);
fit->start = outerStart;
fit->end = innerEnd;
fit->orientation = newbieOrientation;
fit->addAfter = TRUE;
fit->score = fitScore;
return fit;
}

void placeBuriedClones(struct oogClone *cloneList, struct dlList *bargeList, struct hash *ocpHash)
/* Place buried clones in proper position in barges. */
{
struct oogClone *clone, *outerClone;
struct bargeEl *outerBel;
struct barge *barge;
struct dlNode *outCloneStart, *outCloneEnd;
struct dlNode *n1, *n2;

for (clone = cloneList; clone != NULL; clone = clone->next)
    {
    if ((outerBel = clone->bel->enclosed) != NULL)
        {
	struct fitAt *fitList = NULL, *fit;
	outerClone = outerBel->clone;
	if ((barge = outerClone->barge) == NULL)
	    barge = bargeOfOne(outerClone, bargeList);
	uglyf("Trying to place enclosed %s\n", clone->name);
	dumpCloneEnds(barge->cloneEndList, logFile);
	findCloneEndsOnList(barge->cloneEndList, outerClone, &outCloneStart, &outCloneEnd);
	for (n1 = outCloneStart; n1 != outCloneEnd; n1 = n1->next)
	    {
	    for (n2 = n1->next; n2 != outCloneEnd->next; n2 = n2->next)
	        {
		if ((fit = fitEnclosed(barge->cloneEndList->head, 
			outCloneEnd, clone, n1, n2, ocpHash)) != NULL)
		    {
		    dumpFit(fit, clone);
		    slAddHead(&fitList, fit);
		    }
		}
	    }
	slSort(&fitList, fitAtCmpScore);
	if ((fit = fitList) != NULL)
	    {
	    addFittedClone(barge, clone, fit);
	    dumpCloneEnds(barge->cloneEndList, logFile);
	    }
	else
	    warn("Couldn't place enclosed clone %s", clone->name);
	slFreeList(&fitList);
	uglyf("\n");
	}
    }
}

void makeSingletonBarges(struct oogClone *cloneList, struct dlList *bargeList)
/* Make singleton barges around clones not barged up yet. */
{
struct oogClone *clone;
for (clone = cloneList; clone != NULL; clone = clone->next)
    {
    if (clone->barge == NULL)
        {
	bargeOfOne(clone, bargeList);
	}
    }
}

/* Place buried clones in proper position in barges. */
struct dlList *mmBargeList(struct oogClone *cloneList, struct hash *cloneHash,
	struct overlappingClonePair *ocpList, struct hash *ocpHash)
/* Make contigs of overlapping clones (aka barges) using overlap info. 
 * ocpList should be sorted by score. */
{
struct overlappingClonePair *ocp;
struct dlList *bargeList = newDlList();
struct barge *barge, *bargeA, *bargeB;

clonesToBels(cloneList);
markBuriedClones(ocpList);
for (ocp = ocpList; ocp != NULL; ocp = ocp->next)
    {
    ocpDump(ocp, uglyOut);
    ocpDump(ocp, logFile);
    if (ocp->score < 0)
        break;
    if (ocp->a->bel->enclosed || ocp->b->bel->enclosed)
        {
	uglyf("skipping enclosed\n\n");
	continue;
	}
    if (startsWith("uD0", ocp->a->name) && startsWith("uT0", ocp->b->name))
        uglyf("Got a live one!\n");
    bargeA = ocp->a->barge;
    bargeB = ocp->b->barge;
    if (bargeA == NULL && bargeB == NULL)
        {
	barge = bargeOfTwo(ocp, bargeList, ocpHash, ocpList);
	}
    else if (bargeA == NULL)
        {
	addToBarge(bargeB, ocp, ocp->b, ocp->a, ocpHash, ocpList);
	}
    else if (bargeB == NULL)
        {
	addToBarge(bargeA, ocp, ocp->a, ocp->b, ocpHash, ocpList);
	}
    else if (bargeA == bargeB)
        {
	/* This is handled elsewhere. */
	}
    else if (bargeA != bargeB)
        {
	mergeBarges(bargeA, bargeB, ocp, ocpHash, bargeList);
	}
    uglyf("\n");
    }
placeBuriedClones(cloneList, bargeList, ocpHash);
makeSingletonBarges(cloneList, bargeList);
return bargeList;
}

void mm(char *testSet)
/* mm - Map mender. */
{
char fileName[512];
struct oogClone *cloneList = NULL, *clone;
struct hash *cloneHash = NULL;
struct overlappingClonePair *ocpList = NULL, *ocp;
struct hash *ocpHash = NULL;
struct dlList *bargeList = NULL;
struct barge  *barge;
struct dlNode *node;
FILE *f;

/* Set up logging of error and status messages. */
sprintf(fileName, "%s.log", testSet);
logFile = mustOpen(fileName, "w");
pushWarnHandler(warnHandler);

sprintf(fileName, "%s.clones", testSet);
mmLoadClones(fileName, &cloneList, &cloneHash);
printf("Loaded %d clones from %s\n", slCount(cloneList), fileName);

sprintf(fileName, "%s.overlap", testSet);
mmLoadOcp(fileName, cloneHash, &ocpList, &ocpHash);
printf("Loaded %d overlaps from %s\n", slCount(ocpList), fileName);


slSort(&ocpList, ocpCmpScore);
for (ocp = ocpList; ocp != NULL; ocp = ocp->next)
    ocpDump(ocp, logFile);

bargeList = mmBargeList(cloneList, cloneHash, ocpList, ocpHash);

sprintf(fileName, "%s.ends", testSet);
f = mustOpen(fileName, "w");
uglyf("Final barges:\n");
for (node = bargeList->head; !dlEnd(node); node = node->next)
    {
    barge = node->val;
    dumpCloneEnds(barge->cloneEndList, logFile);
    dumpCloneEnds(barge->cloneEndList, uglyOut);
    dumpCloneEnds(barge->cloneEndList, f);
    }
carefulClose(&f);
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 2)
    usage();
mm(argv[1]);

#ifdef SOON
mm("../clean/big");
#endif
return 0;
}
